/*
 * Copyright (c) 2014-2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Tue Feb 18 2014
 */
#include "acoptitrack.h"

#include <sys/select.h>
#include <sys/time.h>
#include <assert.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "optitrack_c_types.h"
#include "codels.h"


/* --- Task publish ----------------------------------------------------- */


/** Codel opti_publish_start of task publish.
 *
 * Triggered by optitrack_start.
 * Yields to optitrack_pause_descr.
 * Throws optitrack_e_nomem.
 */
genom_event
opti_publish_start(optitrack_ids *ids, const genom_context self)
{
  ids->pubdata.bodies = true;
  ids->pubdata.markers = false;
  ids->pubdata.anon = false;

  ids->local_port = 1510;
  ids->server = NULL;
  ids->binfo = calloc(1, sizeof(ids->binfo));
  if (!ids->binfo) return optitrack_e_nomem(self);

  ids->noise.pstddev = 0.;
  ids->noise.qstddev = 0.;

  return optitrack_pause_descr;
}


/** Codel opti_publish_descr of task publish.
 *
 * Triggered by optitrack_descr.
 * Yields to optitrack_pause_descr, optitrack_recv,
 *           optitrack_disconnect.
 * Throws optitrack_e_nomem.
 */
genom_event
opti_publish_descr(optitrack_server_s **server,
                   const genom_context self)
{
  optitrack_e_network_detail n;
  int e;

  if (!server || !*server) return optitrack_pause_descr;

  e = natnet_snd_modeldef((*server)->natnet, n.what, sizeof(n.what));
  if (e) {
    warnx("%s", n.what);
    return optitrack_disconnect;
  }

  return optitrack_recv;
}


/** Codel opti_publish_recv of task publish.
 *
 * Triggered by optitrack_recv.
 * Yields to optitrack_command, optitrack_data, optitrack_recv,
 *           optitrack_disconnect, optitrack_pause_descr.
 * Throws optitrack_e_nomem.
 */
genom_event
opti_publish_recv(const optitrack_server_s *server,
                  const genom_context self)
{
  struct timeval tv = { .tv_sec = 1, .tv_usec = 0 };
  optitrack_e_network_detail d;
  int e;

  if (!server) return optitrack_pause_descr;

  e = natnet_poll(server->natnet, &tv);
  if (e == NATNET_EVERR) return optitrack_disconnect;

  if (e & NATNET_EVCMD) return optitrack_command;
  if (e & NATNET_EVDAT) return optitrack_data;

  /* timeout: send ping, so that if the server was switched from multicast to
   * unicast, it starts sending frame data */
  if (natnet_snd_ping(server->natnet, d.what, sizeof(d.what)))
    return optitrack_disconnect;

  return optitrack_recv;
}

/** Codel opti_publish_command of task publish.
 *
 * Triggered by optitrack_command.
 * Yields to optitrack_descr, optitrack_recv, optitrack_disconnect.
 * Throws optitrack_e_nomem.
 */
genom_event
opti_publish_command(optitrack_server_s **server,
                     optitrack_ids_bw_s *bw,
                     optitrack_ids_ping_ctx *ping,
                     optitrack_bodies_info_s **binfo,
                     const optitrack_log_s *log,
                     const optitrack_ids_pubdata_s *pubdata,
                     const optitrack_ids_noise_s *noise,
                     const optitrack_bodies *bodies,
                     const optitrack_markers *markers,
                     const genom_context self)
{
  optitrack_e_network_detail d;
  genom_event e;
  size_t sz;
  int s;

  s = natnet_cmd_recv((*server)->natnet, &sz, d.what, sizeof(d.what));
  switch (s) {
    case NATNET_MSGPING:
      natnet_rcv_ping((*server)->natnet,
                      ping->response, sizeof(ping->response));
      ping->id++;
      break;

    case NATNET_MSGMODELDEF:
      (*binfo)->descr = natnet_rcv_modeldef((*server)->natnet);
      if ((*binfo)->descr) {
        e = opti_create_ports((*binfo)->descr, pubdata, bodies, markers, self);
        if (e != genom_ok) return optitrack_disconnect;
      }
      break;

    case NATNET_MSGFRAMEOFDATA: {
      /* in unicast mode, frame data is sent to the command socket */
      const struct natnet_frame *f;

      f = natnet_rcv_frameofdata((*server)->natnet);
      if (f) {
        if (f->refresh)
          (*binfo)->descr = NULL; /* refetch descriptions */
        else
          opti_update_ports(f, *server, *binfo, pubdata, noise, bodies, markers,
                            log?log->f:NULL, self);
      }
      break;
    }

    case NATNET_MSGERR:
      warnx("%s", d.what);
      return optitrack_disconnect;
  }

  /* update bandwidth */
  {
    struct timeval tv;
    double dt;

    gettimeofday(&tv, NULL);
    dt = tv.tv_sec - bw->ts.sec + (tv.tv_usec - bw->ts.nsec/1000.)*1e-6;
    bw->ts.sec = tv.tv_sec;
    bw->ts.nsec = tv.tv_usec * 1000;
    bw->cur = sz/dt/(1<<20);
    bw->avg = (127 * bw->avg + bw->cur)/128;
  }

  if (!(*binfo)->descr) return optitrack_descr;
  return optitrack_recv;
}


/** Codel opti_publish_data of task publish.
 *
 * Triggered by optitrack_data.
 * Yields to optitrack_descr, optitrack_recv.
 * Throws optitrack_e_nomem.
 */
genom_event
opti_publish_data(optitrack_server_s **server, optitrack_ids_bw_s *bw,
                  optitrack_bodies_info_s **binfo,
                  const optitrack_log_s *log,
                  const optitrack_ids_pubdata_s *pubdata,
                  const optitrack_ids_noise_s *noise,
                  const optitrack_bodies *bodies,
                  const optitrack_markers *markers,
                  const genom_context self)
{
  optitrack_e_network_detail d;
  size_t sz;
  int e;

  e = natnet_data_recv((*server)->natnet, &sz, d.what, sizeof(d.what));
  switch (e) {
    case NATNET_MSGFRAMEOFDATA: {
      const struct natnet_frame *f;

      f = natnet_rcv_frameofdata((*server)->natnet);
      if (f) {
        if (f->refresh)
          (*binfo)->descr = NULL; /* refetch descriptions */
        else
          opti_update_ports(f, *server, *binfo, pubdata, noise, bodies, markers,
                            log?log->f:NULL, self);
      }
      break;
    }

    case NATNET_MSGERR:
      return optitrack_disconnect;
  }

  /* update bandwidth */
  {
    struct timeval tv;
    double dt;

    gettimeofday(&tv, NULL);
    dt = tv.tv_sec - bw->ts.sec + (tv.tv_usec - bw->ts.nsec/1000.)*1e-6;
    bw->ts.sec = tv.tv_sec;
    bw->ts.nsec = tv.tv_usec * 1000;
    bw->cur = sz/dt/(1<<20);
    bw->avg = (8191 * bw->avg + bw->cur)/8192;
  }

  if (!(*binfo)->descr) return optitrack_descr;
  return optitrack_recv;
}


/** Codel opti_publish_disconnect of task publish.
 *
 * Triggered by optitrack_disconnect.
 * Yields to optitrack_pause_descr.
 * Throws optitrack_e_nomem.
 */
genom_event
opti_publish_disconnect(optitrack_server_s **server,
                        const genom_context self)
{
  if (server && *server) {
    natnet_disconnect((*server)->natnet);
    free(*server);
    *server = NULL;
  }

  return optitrack_pause_descr;
}


/** Codel opti_publish_stop of task publish.
 *
 * Triggered by optitrack_stop.
 * Yields to optitrack_ether.
 * Throws optitrack_e_nomem.
 */
genom_event
opti_publish_stop(optitrack_server_s **server,
                  optitrack_bodies_info_s **binfo,
                  const genom_context self)
{
  opti_publish_disconnect(server, self);

  if (binfo && *binfo) {
    free(*binfo);
  }

  return optitrack_ether;
}


/* --- Activity ping ---------------------------------------------------- */

/** Codel opti_ping_start of activity ping.
 *
 * Triggered by optitrack_start.
 * Yields to optitrack_send.
 * Throws optitrack_e_nomem, optitrack_e_connection,
 *        optitrack_e_timeout, optitrack_e_network.
 */
genom_event
opti_ping_start(const optitrack_server_s *server,
                const optitrack_ids_ping_ctx *ping, int32_t *id,
                double *ts, int32_t *retries,
                const genom_context self)
{
  struct timeval tv;
  if (!server || !server->natnet) return optitrack_e_connection(self);

  gettimeofday(&tv, NULL);
  *ts = tv.tv_sec + 1e-6 * tv.tv_usec;
  *id = ping->id + 1;
  *retries = 0;
  return optitrack_send;
}

/** Codel opti_ping_send of activity ping.
 *
 * Triggered by optitrack_send.
 * Yields to optitrack_pause_sleep.
 * Throws optitrack_e_nomem, optitrack_e_connection,
 *        optitrack_e_timeout, optitrack_e_network.
 */
genom_event
opti_ping_send(const optitrack_server_s *server,
               const genom_context self)
{
  optitrack_e_network_detail d;
  if (!server || !server->natnet) return optitrack_e_connection(self);

  if (natnet_snd_ping(server->natnet, d.what, sizeof(d.what)))
    return optitrack_e_network(&d, self);
  return optitrack_pause_sleep;
}

/** Codel opti_ping_sleep of activity ping.
 *
 * Triggered by optitrack_sleep.
 * Yields to optitrack_send, optitrack_stop.
 * Throws optitrack_e_nomem, optitrack_e_connection,
 *        optitrack_e_timeout, optitrack_e_network.
 */
genom_event
opti_ping_sleep(const optitrack_ids_ping_ctx *ping, int32_t id,
                int32_t *retries, const genom_context self)
{
  if (ping->id < id) {
    if (++(*retries) < 7) return optitrack_send;
    return optitrack_e_timeout(self);
  }
  return optitrack_stop;
}

/** Codel opti_ping_stop of activity ping.
 *
 * Triggered by optitrack_stop.
 * Yields to optitrack_ether.
 * Throws optitrack_e_nomem, optitrack_e_connection,
 *        optitrack_e_timeout, optitrack_e_network.
 */
genom_event
opti_ping_stop(const optitrack_ids_ping_ctx *ping, double ts,
               double *rtt, char info[64], const genom_context self)
{
  struct timeval tv;

  gettimeofday(&tv, NULL);

  *rtt = tv.tv_sec + 1e-6 * tv.tv_usec - ts;
  snprintf(info, 64/*sizeof(info), grr C arrays ...*/, "%s", ping->response);
  return optitrack_ether;
}


/* --- Activity connect ------------------------------------------------- */

/** Codel opti_connect_start of activity connect.
 *
 * Triggered by optitrack_start.
 * Yields to optitrack_ether.
 * Throws optitrack_e_nomem, optitrack_e_addrinfo.
 */
genom_event
opti_connect_start(uint16_t local_port, const char host[128],
                   const char host_port[128], const char mcast[128],
                   const char mcast_port[128],
                   optitrack_server_s **server,
                   const genom_context self)
{
  optitrack_e_addrinfo_detail d;

  /* private context */
  opti_publish_disconnect(server, self);
  *server = calloc(1, sizeof(**server));
  if (!*server) return optitrack_e_nomem(self);

  /* setup network */
  (*server)->natnet = natnet_connect(local_port, host, host_port,
                                     mcast, mcast_port, d.what, sizeof(d.what));
  if (!(*server)->natnet) {
    warnx("%s", d.what);
    free(*server);
    *server = NULL;
    return optitrack_e_addrinfo(&d, self);
  }

  return optitrack_ether;
}


/* --- Activity disconnect ---------------------------------------------- */

/** Codel opti_disconnect of activity disconnect.
 *
 * Triggered by optitrack_start.
 * Yields to optitrack_ether.
 * Throws optitrack_e_nomem.
 */
genom_event
opti_disconnect(optitrack_server_s **server, const genom_context self)
{
  opti_publish_disconnect(server, self);
  return optitrack_ether;
}
